import math
from quicktions import Fraction


class ArithmeticProgression(object):
    def __init__(self, a1=None, an=None, n=None, d=None, s=None):
        self._a1 = None
        self._an = None
        self._n = None
        self._d = None
        self._s = None
        self._current = None
        self._index = None

        self.a1 = a1
        self.an = an
        self.n = n
        self.d = d
        self.s = s

    def _check_args(self):

        if len([v for v in self._parameters_dict.values() if v is not None]) > 2:
            err = 'attribute cannot be set. Three parameters are already set. ArithmeticProgression is already created!'
            raise AttributeError(err)

    def _calculate_a1(self):
        if self._d is None:
            self._a1 = (2. * self.s / self.n) - self.an
        elif self._s is None:
            self._a1 = self.an - ((self.n - 1) * self.d)

    def _calculate_an(self):
        if self._s is None:
            self._an = self.a1 + (self.n - 1) * self.d
        elif self._d is None:
            self._an = (2. * self.s / self.n) - self.a1

    def _calculate_n(self):
        if self._s is None:
            self._n = ((self.an - self.a1) / self.d) + 1
        elif self._d is None:
            self._n = 2. * self.s / (self.a1 + self.an)
        self._n = math.floor(self._n)

    def _calculate_d(self):

        if self._a1 is None:
            self._calculate_a1()
            self._d = Fraction(Fraction(self.an - self.a1), Fraction(self.n - 1))
        elif self._an is None:
            self._d = Fraction(Fraction((self.s - (self.n * self.a1)) * 2), Fraction((self.n - 1) * self.n))
        elif self._n is None:
            self._calculate_n()
            self._d = Fraction(Fraction(self.an - self.a1), Fraction(self.n - 1))
        else:
            self._d = Fraction(Fraction(self.an - self.a1), Fraction(self.n - 1))

    def _calculate_s(self):
        if self._a1 is None:
            self._calculate_a1()
            self._s = self.n * (self.a1 + self.an) / 2.
        elif self._an is None:
            self._s = self.n * self.a1 + ((self.n - 1) * self.n / 2) * self.d
        elif self._n is None:
            self._calculate_n()
            self._s = self.n * (self.a1 + self.an) / 2.
        else:
            self._s = self.n * (self.a1 + self.an) / 2.

    @property
    def _parameters_dict(self):
        return {'a': self._a1, 'an': self._an, 'n': self._n, 'd': self._d, 's': self._s}

    @property
    def parameters_dict(self):
        return {'a': self.a1, 'an': self.an, 'n': self.n, 'd': self.d, 's': self.s}

    @property
    def a1(self):
        if self._a1 is None:
            self._calculate_a1()
        return self._a1

    @a1.setter
    def a1(self, value):
        if value is not None:
            self._check_args()
        self._a1 = value

    @property
    def an(self):
        if self._an is None:
            self._calculate_an()
        return self._an

    @an.setter
    def an(self, value):
        if value is not None:
            self._check_args()
        self._an = value

    @property
    def n(self):
        if self._n is None:
            self._calculate_n()
        return self._n

    @n.setter
    def n(self, value):
        if value is not None:
            self._check_args()
        self._n = value

    @property
    def s(self):
        if self._s is None:
            self._calculate_s()
        return self._s

    @s.setter
    def s(self, value):
        if value is not None:
            self._check_args()
            if self._d is not None:
                err = 'you cannot set both d an s!'
                raise AttributeError(err)
        self._s = value

    @property
    def d(self):
        if self._d is None:
            self._calculate_d()
        return self._d

    @d.setter
    def d(self, value):
        if value is not None:
            self._check_args()
            if self._s is not None:
                err = 'you cannot set both d an s!'
                raise AttributeError(err)
        self._d = value

    def __iter__(self):
        return self

    def __next__(self):
        parameters = [self.a1, self.an, self.n, self.s, self.d]

        if len([v for v in parameters if v is not None]) < 5:
            err = 'Not enough parameter set to create an arithmetic progression. 3 parameters should be set first (s ' \
                  'and d cannot be set together) '
            raise Exception(err)

        if self._current is None:
            self._current = self.a1
            self._index = 0
        else:
            self._index += 1
            self._current += self.d

        if self._index < self.n:
            return self._current
        else:
            raise StopIteration()

    @property
    def index(self):
        return self._index

    def rest(self):
        self._current = None



from unittest import TestCase

from musurgia.agrandom import AGRandom


class Test(TestCase):
    def test_1(self):
        pool = [1, 3, 2, 4, 5, 1, 1]
        r = AGRandom(pool=pool, periodicity=2, seed=20)
        output = [r.__next__() for i in range(20)]
        result = [2, 3, 1, 5, 2, 1, 4, 2, 3, 4, 5, 2, 3, 4, 1, 5, 4, 1, 2, 5]
        self.assertEqual(output, result)

    def test_2(self):
        pool = [1, 3, 2, 4, 5, 1, 1, 1, 1, 1, 1]
        r = AGRandom(pool=pool, periodicity=0, seed=20)
        result = [1, 2, 3, 4, 5]
        self.assertEqual(r.pool, result)

    def test_3(self):
        pool = [1, 3, 2, 4, 5, 6]
        r = AGRandom(pool=pool, periodicity=0, seed=20)
        output = [r.__next__() for i in range(20)]
        result = [6, 6, 2, 3, 6, 6, 1, 3, 5, 2, 1, 4, 4, 1, 1, 2, 3, 4, 5, 4]
        self.assertEqual(output, result)



import itertools
import os
from math import ceil, floor

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treechord import TreeChord
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from prettytable import PrettyTable
from quicktions import Fraction

from musurgia import basic_functions, scaledvalues
from musurgia.fractaltree.fractaltree import FractalTree
from musurgia.fractaltree.midigenerators import RelativeMidi, MidiGenerator
from musurgia.permutation import permute
from musurgia.quantize import get_quantized_values
from musurgia.timing import Timing


class FractalMusicException(Exception):
    def __init__(self, *args):
        super().__init__(*args)


class SetDurationFirstException(FractalMusicException):
    def __init__(self, *args):
        super().__init__('set duration first!', *args)


class SetTempoFirstException(FractalMusicException):
    def __init__(self, *args):
        super().__init__('set tempo first!', *args)


class TempoIsAlreadySet(FractalMusicException):
    def __init__(self, *args):
        super().__init__('FractalMusic().tempo can only be set once', *args)


class ChildTempoIsAlreadySet(FractalMusicException):
    def __init__(self, *args):
        super().__init__('FractalMusic().tempo of parent can not be set after setting tempo of child', *args)


class MergeException(FractalMusicException):
    def __init__(self, *args):
        super().__init__(*args)


class MergeTempoException(MergeException):
    def __init__(self, tempo, *args):
        super().__init__('nodes to merge must have the same tempo {}'.format(tempo), *args)


class FractalMusic(FractalTree):
    def __init__(self,
                 duration=None,
                 midi_generator=None,
                 tempo=None,
                 quarter_duration=None,
                 tree_directions=None,
                 permute_directions=True,
                 *args, **kwargs):

        # super().__init__(value=duration, *args, **kwargs)
        super().__init__(*args, **kwargs)
        self._midi_value = None
        self._chord = None
        self._midi_generator = None
        self._children_generated_midis = None
        self._tree_directions = None
        self._permute_directions = True
        self._tempo = None

        self.midi_generator = midi_generator
        if duration:
            self.duration = duration

        self.tempo = tempo
        self.permute_directions = permute_directions
        self.tree_directions = tree_directions

        self.quarter_duration = quarter_duration

    @property
    def duration(self):
        return self.value

    @duration.setter
    def duration(self, val):
        self.value = val

    @property
    def quarter_duration(self):
        if not self.tempo:
            raise SetTempoFirstException()
        return self.duration * self.tempo / 60.

    @quarter_duration.setter
    def quarter_duration(self, val):
        if val is not None:
            if not self.tempo:
                raise SetTempoFirstException()
            try:
                self.duration = Fraction(val * 60, self.tempo)
            except TypeError:
                self.duration = Fraction(Fraction(val * 60), Fraction(self.tempo))

    def _child_has_tempo(self):
        children_tempi = [child.tempo for child in self.get_children() if child.tempo]
        if children_tempi:
            return True
        else:
            return False

    @property
    def tempo(self):
        return self._tempo

    @tempo.setter
    def tempo(self, val):
        if self._tempo:
            raise TempoIsAlreadySet()
        if self._child_has_tempo():
            raise ChildTempoIsAlreadySet()
        self._tempo = val
        for child in self.get_children():
            child.tempo = val

    def set_non_tempi(self, val=60):
        try:
            self.tempo = val
        except TempoIsAlreadySet:
            pass
        except ChildTempoIsAlreadySet:
            for child in self.get_children():
                child.set_non_tempi(val=val)

    def find_best_tempo(self, min_tempo=40, max_tempo=144):
        min_quarter_duration = ceil(Timing(duration=self.duration, tempo=min_tempo).quarter_duration)
        max_quarter_duration = floor(Timing(duration=self.duration, tempo=max_tempo).quarter_duration)
        tempi = [Timing(duration=self.duration, quarter_duration=x).tempo for x in
                 range(min_quarter_duration, max_quarter_duration + 1)]
        tempo_deltas = [abs(round(tempo) - tempo) for tempo in tempi]
        min_delta = min(tempo_deltas)
        index = tempo_deltas.index(min_delta)
        best_tempo = int(round(tempi[index]))
        return best_tempo

    @property
    def quarter_position_in_tree(self):
        return self.position_in_tree * self.tempo / 60

    @property
    def permute_directions(self):
        return self._permute_directions

    @permute_directions.setter
    def permute_directions(self, value):
        if not isinstance(value, bool):
            raise TypeError()
        self._permute_directions = value

    @property
    def tree_directions(self):
        return self._tree_directions

    # todo: setting tree_direction after setting midi_range has no effect
    @tree_directions.setter
    def tree_directions(self, val):
        if val is None:
            val = [1, -1]
        length = len(self.permutation_order)
        if len(val) > length:
            val = val[:length]
        elif len(val) < length:
            output = []
            cycled = itertools.cycle(val)
            for i in range(length):
                output.append(cycled.__next__())
            val = output
        self._tree_directions = val

    @property
    def midi_generator(self):
        if self._midi_generator is None:
            try:
                if not self.duration:
                    raise SetDurationFirstException()
                self._midi_generator = RelativeMidi(midi_range=[71, 71], proportions=self.children_fractal_values,
                                                    directions=None)
                self._midi_generator.node = self
            except AttributeError:
                raise Exception('midi_generator is None and cannot be set:')
        else:
            self._midi_generator.node = self

        if isinstance(self._midi_generator, RelativeMidi):

            if not self._midi_generator._directions:
                tree_directions = self.get_root().tree_directions

                if self.permute_directions:
                    directions = permute(tree_directions, self.permutation_order)
                else:
                    directions = tree_directions

                self._midi_generator.directions = directions

            if self._midi_generator.midi_range is None:
                if self.is_root:
                    self._midi_generator.midi_range = [71]
                else:
                    self._midi_generator.midi_range = self.auto_midi_range
                    self._midi_generator._auto_ranged = True

        return self._midi_generator

    @midi_generator.setter
    def midi_generator(self, value):
        if value is not None:
            if not (isinstance(value, MidiGenerator)):
                err = 'midi_generator can only be an instance of subclasses of MidiGenerator or None. None=RelativeMidi(midi_range=self.midi_range, proportions=permute(self.fractal_proportions, self.fractal.permutation_order)'
                raise TypeError(err)

        self._midi_generator = value
        if self._midi_generator:
            self._midi_generator.node = self

    @property
    def _midi_iterator(self):
        mg = self.midi_generator
        return mg.iterator

    @_midi_iterator.setter
    def _midi_iterator(self, value):
        raise AttributeError('_midi_iterator cannot be set directly. Use midi-generator!')

    @property
    def children_generated_midis(self):
        if self._children_generated_midis is None:
            self._children_generated_midis = []
            self._children_generated_midis = list(self._midi_iterator)

        return self._children_generated_midis

    @property
    def _children_midis(self):
        return [child.midi_value for child in self.get_children()]

    @property
    def midi_value(self):
        if self._midi_value is None:
            if self.is_root:
                self._midi_value = 71
            else:
                self._midi_value = self.up.children_generated_midis[self.up.get_children().index(self)]
        return self._midi_value

    @midi_value.setter
    def midi_value(self, value):
        if value is not None:
            if isinstance(value, int) or isinstance(value, float):
                if value >= 18 or value == 0:
                    self._midi_value = value
                else:
                    raise ValueError('midi can be 0 and greater than 18')
            else:
                raise TypeError('midi can only be int, float or None not {}'.format(type(value)))
        else:
            self._midi_value = value

    @property
    def auto_midi_range(self):
        if self.is_root:
            raise AttributeError('root has no auto_midi_range')

        parent_midis = self.up.children_generated_midis
        self_index = self.up.get_children().index(self)
        auto_range = parent_midis[self_index:self_index + 2]
        return auto_range

    def get_choral_midis(self, range_factor=1, direction=None, last=False):
        range_ = (self.midi_generator.midi_range[1] - self.midi_generator.midi_range[0]) * range_factor
        self_direction = range_ / abs(range_)
        if direction is None or direction == 0:
            pass
        elif direction == 1:
            range_ = abs(range_)
        elif direction == -1:
            range_ = -abs(range_)

        children_midis = self.children_generated_midis
        choral_midis = []

        step = 2 / self.midi_generator.microtone
        if self_direction > 0:
            scale = scaledvalues.ScaledValues(min(children_midis), max(children_midis), self.midi_value,
                                              self.midi_value + range_,
                                              step=step)
        else:
            scale = scaledvalues.ScaledValues(max(children_midis), min(children_midis), self.midi_value,
                                              self.midi_value + range_,
                                              step=step)
        if not last:
            min_midi = min(children_midis)
            max_midi = max(children_midis)
            if self.midi_value == min_midi:
                children_midis.remove(max_midi)
            else:
                children_midis.remove(min_midi)
        for midi in children_midis:
            new_midi = scale(midi)
            choral_midis.append(new_midi)

        choral_midis = sorted(choral_midis)
        choral_midis = list(set(choral_midis))
        return choral_midis

    def get_choral(self, range_factor=1, direction=None, last=False):
        choral_midis = self.get_choral_midis(range_factor=range_factor, direction=direction, last=last)
        return TreeChord(quarter_duration=self.quarter_duration, midis=choral_midis)

    @property
    def chord(self):
        if self._chord is None:  # or self._note.midis!=[self.midi] or self._note.duration!=self.duration:
            self._chord = TreeChord(quarter_duration=self.quarter_duration, midis=[self.midi_value])
        else:
            self._chord.quarter_duration = self.quarter_duration

        return self._chord

    def set_chord(self, chord):
        self._chord = chord

    def get_simple_format(self, layer=None):
        if layer is None:
            layer = self.get_farthest_leaf().get_distance()
        self.get_layer(layer=layer)

        simple_format = SimpleFormat()
        for chord in [node.chord for node in basic_functions.flatten(self.get_layer(layer))]:
            copied_chord = chord.__deepcopy__()
            simple_format.add_chord(copied_chord)

        return simple_format

    def split(self, *proportions):
        if hasattr(proportions[0], '__iter__'):
            proportions = proportions[0]

        proportions = [Fraction(prop) for prop in proportions]

        for prop in proportions:
            duration = self.quarter_duration * prop / sum(proportions)
            new_node = self.copy()
            new_node.multi = self.multi
            new_node.quarter_duration = duration
            new_node._fractal_order = self.fractal_order
            new_node.midi_value = self.midi_value
            new_node.midi_generator.midi_range = self.midi_generator.midi_range
            if self._chord:
                new_node.set_chord(self.chord.__deepcopy__())
            self.add_child(new_node)

        return self.get_children()

    def _check_merge_nodes(self, nodes):
        tempo = set([node.tempo for node in nodes])
        if len(tempo) != 1:
            raise MergeTempoException(tempo)

    # def merge_children(self, *lengths):
    #     super().merge_children(*lengths)

    def _reset_durations(self):
        for node in self.traverse():
            node._value = None

    def _refill_duration(self):
        if not self._value:
            self._value = sum([child._refill_duration() for child in self.get_children()])
        else:
            return self._value

    def quantize_leaves(self, grid_size):

        #     quantizing quarter_durations!
        leaves = list(self.traverse_leaves())
        durations = [leaf.quarter_duration for leaf in leaves]
        quantized_durations = get_quantized_values(durations, grid_size)
        if quantized_durations:
            self._reset_durations()
        for leaf, quantized_duration in zip(leaves, quantized_durations):
            leaf.quarter_duration = quantized_duration
        self._refill_duration()

    def round_leaves(self):

        #     quantizing quarter_durations!
        leaves = list(self.traverse_leaves())
        rounded_quarter_durations = [round(leaf.quarter_duration) for leaf in leaves]
        if rounded_quarter_durations:
            self._reset_durations()
        for leaf, rounded_quarter_duration in zip(leaves, rounded_quarter_durations):
            leaf.quarter_duration = rounded_quarter_duration
        self._refill_duration()

    def change_midis(self):
        if not isinstance(self._midi_generator, RelativeMidi):
            raise TypeError(
                'set_reduced_auto_ranges can only be applied to FractalMusic nodes with RelativeMidi as midi_generator')
        else:

            # print(self.children_generated_midis)
            directions = self.midi_generator.directions
            midi_range = self.midi_generator.midi_range
            microtone = self.midi_generator.microtone

            # children = iter(self.get_children())
            # old_generated_midis = self.children_generated_midis
            # print("old_generated_midis", old_generated_midis)
            # self._children_generated_midis = [node.midi_value for node in self.get_children()]
            # self._children_generated_midis.append(old_generated_midis[-1])
            # print("new_generated_midis", old_generated_midis)

            self.midi_generator = None
            self.midi_generator.proportions = None
            self.midi_generator.midi_range = midi_range
            self.midi_generator.microtone = microtone
            self.midi_generator.directions = directions

            self._children_generated_midis = None

            for child in self.get_children():
                child.midi_value = None
                child._chord = None

            for child in self.get_children():
                if isinstance(child.midi_generator, RelativeMidi):
                    child.midi_generator.midi_range = None
                    child.midi_generator.directions = None
                    child.midi_generator.proportions = None
                    child.midi_generator._iterator = None
                    child._children_generated_midis = None

    def get_score_template(self):
        score = TreeScoreTimewise()
        score.tuplet_line_width = 2.4
        score.page_style.orientation = 'landscape'
        score.page_style.system_distance = 180
        score.page_style.staff_distance = 150
        score.page_style.top_system_distance = 150
        score.page_style.bottom_margin = 100

        score.add_title('module: {}'.format(self.__name__))
        clock = Timing.get_clock(round(self.duration), mode='msreduced')
        score.add_subtitle(
            'duration: {}'.format(clock))
        score.accidental_mode = 'modern'
        return score

    def get_score(self, score=None, layer_number=None, show_fractal_orders=False, show_midis=None,
                  barline='light-heavy',
                  show_metronome=True):
        if not score:
            score = self.get_score_template()
        score.set_time_signatures(durations=self.quarter_duration)
        if show_fractal_orders:
            for node in self.traverse():
                node.chord.add_lyric(node.fractal_order)

        if show_midis:
            for node in self.traverse():
                midi_value = node.midi_value
                if int(midi_value) == midi_value:
                    midi_value = int(midi_value)
                node.chord.add_words(midi_value, enclosure='none', relative_y=10)

        def layer_to_score(layer_number, part_number):
            try:
                sf = self.get_simple_format(layer_number)
                sf.auto_clef()
                v = sf.to_stream_voice(1)
                v.add_to_score(score, 1, part_number)
            except ValueError:
                print('module {}: number_of_layers={}: getting layer {} not possible'.format(self.__name__,
                                                                                             self.number_of_layers,
                                                                                             i))

        if layer_number is None:
            for i, j in zip(range(self.number_of_layers + 1), range(1, self.number_of_layers + 2)):
                layer_to_score(i, j)

        elif hasattr(layer_number, '__iter__'):
            for i, j in zip(layer_number, range(1, len(layer_number) + 1)):
                layer_to_score(i, j)
        else:
            layer_to_score(layer_number, 1)

        if show_metronome:
            score.get_measure(1).get_part(1).add_metronome(per_minute=self.tempo, relative_y=40)
        score.get_measure(-1).set_barline_style(barline)
        return score

    def get_root_score(self, score=None, layer_number=None, show_fractal_orders=False, show_midis=False
                       # , show_positions=False
                       ):

        if not score:
            score = self.get_score_template()

        try:
            return self.get_score(score=score, layer_number=layer_number, show_fractal_orders=show_fractal_orders,
                                  show_midis=show_midis)
        except SetTempoFirstException as err:
            if not self.get_children():
                raise err
            else:
                for child in self.get_children():
                    score.extend(
                        child.get_score(layer_number=layer_number, show_fractal_orders=show_fractal_orders,
                                        show_midis=show_midis, barline='light-light'))

        # if set_time_signatures:
        #     if layer_number == 0 or not self.get_children():
        #         score.set_time_signatures(durations=self.quarter_duration, barline_style='light-light', times=times)
        #     else:
        #         durations = [child.quarter_duration for child in self.get_children() if child.quarter_duration != 0]
        #         score.set_time_signatures(
        #             durations=durations, barline_style='light-light', times=times)
        # else:

        score.accidental_mode = 'modern'

        # if show_positions:
        #     for node in self.get_children()[1:]:
        #         position_in_tree = float(node.position_in_tree)
        #         quarter_position_in_tree = position_in_tree * self.tempo / 60
        #         position_in_score = quarter_position_in_tree * 60 / self.score_tempo
        #         node.chord.add_words(
        #             Timing.get_clock(time=round(position_in_score, 1), mode='msreduced'),
        #             enclosure='rectangle', relative_y=40)
        score.get_measure(-1).set_barline_style('light-heavy')
        return score

    # get_score = get_fractal_score

    def write_infos(self, file_name):
        os.system('touch ' + file_name)
        file = open(file_name, 'w')
        x = PrettyTable()

        x.field_names = ["name", "f_o", "quarters", "duration", "midi", "perm_ord", "multi", "directions",
                         "midi_range", "childr_midis"]

        for node in self.traverse():
            x.add_row([node.__name__, node.fractal_order, round(float(node.quarter_duration), 2),
                       Timing(quarter_duration=float(node.quarter_duration), tempo=self.tempo).clock,
                       node.midi_value,
                       node.permutation_order, node.multi, node.midi_generator.directions,
                       node.midi_generator.midi_range, node.children_generated_midis]
                      )

        if self.is_root:
            file.write('name: root')
        else:
            file.write('name: {} fo: {}'.format(self.__name__, self.fractal_order))
        file.write('\n')
        file.write(
            'quarter_duration: {}, tempo: {}, duration: {}'.format(round(float(self.quarter_duration), 2),
                                                                   self.tempo,
                                                                   round(float(self.duration), 2)))
        file.write('\n')
        file.write('tree directions: {}'.format(self.tree_directions))
        file.write('\n')
        file.write('\n')
        file.write(x.get_string())
        file.close()

    def copy(self):
        copied_midi_generator = self.midi_generator.copy()

        if isinstance(copied_midi_generator, RelativeMidi):
            copied_midi_generator.midi_range = None
            copied_midi_generator._directions = None
            copied_midi_generator.proportions = None
            copied_midi_generator._iterator = None

        copied = super().copy()
        copied.tempo = self.tempo
        copied.midi_generator = copied_midi_generator
        copied.tree_directions = self.tree_directions
        copied.permute_directions = self.permute_directions

        if self.fertile is False:
            copied._midi_value = self._midi_value
        return copied

    def reduce_children(self, condition):
        super().reduce_children(condition)

        try:
            self.change_midis()
        except TypeError:
            pass

    def __deepcopy__(self, memodict={}):
        copied = super().__deepcopy__()
        copied.tempo = self.tempo
        if self._midi_generator is not None:
            copied._midi_generator = self._midi_generator.__deepcopy__()

        if self.fertile is False:
            copied._midi_value = self._midi_value
        return copied



import os
from unittest import TestCase

from musurgia.agpdf.pdf import Pdf
from quicktions import Fraction

from musurgia.testcomparefiles import TestCompareFiles
from musurgia.timeline.timeline import TimeLine

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):
    def test_1(self):
        pdf_path = path + '_test_1.pdf'
        pdf = Pdf(orientation='portrait')
        tl = TimeLine(length=50)
        printable = (pdf.w - pdf.r_margin - pdf.l_margin)
        tl.unit = Fraction(Fraction(printable), Fraction(20))
        tl.show_interval = 10
        tl.ruler.draw(pdf)
        pdf.write(pdf_path)
        TestCompareFiles().assertTemplate(file_path=pdf_path)

    def test_2(self):
        pdf_path = path + '_test_2.pdf'
        pdf = Pdf(orientation='portrait')
        tl = TimeLine(length=300)
        printable = (pdf.w - pdf.r_margin - pdf.l_margin)
        tl.unit = Fraction(Fraction(printable), Fraction(20))
        tl.show_interval = 10
        tl.ruler.draw(pdf)
        pdf.write(pdf_path)
        TestCompareFiles().assertTemplate(file_path=pdf_path)




from musurgia.agpdf.drawobject import DrawObject
from musurgia.agpdf.labeled import Labeled


class MarkLine(DrawObject, Labeled):
    def __init__(self, height=3, thickness=1, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._height = None
        self.height = height
        self.position = None
        self.thickness = thickness

    @property
    def height(self):
        return self._height

    @height.setter
    def height(self, val):
        self._height = val

    def get_relative_x2(self):
        return self.relative_x

    def get_relative_y2(self):
        return self.relative_y + self.height

    def draw(self, pdf):
        x1 = self.relative_x + pdf.x
        x2 = x1
        y1 = self.relative_y + pdf.y - self.height / 2
        y2 = y1 + self.height
        if self.show:
            for i in range(self.thickness):
                grid = 0.1
                pdf.line(x1=x1 + i * grid, y1=y1, x2=x2 + i * grid, y2=y2)
        self.position = (x1, y2)
        for text_label in self.text_labels:
            text_label.draw(pdf)
        # pdf.x = x2
        # pdf.y = y2



import os
from unittest import TestCase

from musurgia.agpdf.line import Line
from musurgia.agpdf.pdf import Pdf
from musurgia.agpdf.textlabel import TextLabel
from musurgia.testcomparefiles import TestCompareFiles

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def test_1(self):
        pdf_path = path + '_test_1.pdf'
        pdf = Pdf()
        line = Line(length=60, relative_x=70)
        line.end_mark_line.show = True
        line.draw(pdf=pdf)
        pdf.write(pdf_path)
        TestCompareFiles().assertTemplate(file_path=pdf_path)

    def test_2(self):
        pdf_path = path + '_test_2.pdf'
        pdf = Pdf()
        line_1 = Line(length=10)
        line_2 = Line(length=20, relative_x=10)
        line_2.end_mark_line.show = True
        line_1.draw_with_break(pdf)
        line_2.draw_with_break(pdf)
        pdf.write(pdf_path)
        TestCompareFiles().assertTemplate(file_path=pdf_path)

    def test_3(self):
        pdf_path = path + '_test_3.pdf'
        pdf = Pdf()
        line_1 = Line(length=10, relative_x=20)
        line_2 = Line(length=20, relative_x=200)
        line_2.end_mark_line.show = True
        line_1.draw_with_break(pdf)
        line_2.draw_with_break(pdf)
        pdf.write(pdf_path)

    def test_4(self):
        pdf_path = path + '_test_4.pdf'
        pdf = Pdf(orientation='landscape')
        line_1 = Line(length=10)
        line_3 = Line(length=20, relative_y=190)
        line_1.draw_with_break(pdf)
        line_3.draw_with_break(pdf)
        pdf.write(pdf_path)
        TestCompareFiles().assertTemplate(file_path=pdf_path)

    def test_5(self):
        pdf_path = path + '_test_5.pdf'
        pdf = Pdf()
        line = Line(length=60, relative_x=70)
        line.end_mark_line.show = True
        line.add_text_label(TextLabel('bla', font_size=8))
        line.draw(pdf=pdf)
        pdf.write(pdf_path)
        TestCompareFiles().assertTemplate(file_path=pdf_path)

    def test_6(self):
        pdf_path = path + '_test_6.pdf'
        pdf = Pdf()
        line = Line(length=60, relative_x=70)
        line.end_mark_line.show = True
        TextLabel.DEFAULT_FONT_SIZE = 8

        line.end_mark_line.add_text_label(TextLabel(3))
        line.end_mark_line.add_text_label(TextLabel(4, relative_y=-4))
        line.end_mark_line.add_text_label(TextLabel(5, relative_y=-6))
        line.add_text_label(TextLabel('bla'))
        line.add_text_label(TextLabel('bla bla', relative_y=-4))
        line.draw(pdf=pdf)
        pdf.write(pdf_path)
        TestCompareFiles().assertTemplate(file_path=pdf_path)
        TextLabel.DEFAULT_FONT_SIZE = 10

    def test_7(self):
        pdf_path = path + '_test_7.pdf'
        pdf = Pdf()
        line = Line(length=20, relative_x=0, factor=5)
        line.end_mark_line.show = True
        line.draw(pdf=pdf)
        pdf.write(pdf_path)
        TestCompareFiles().assertTemplate(file_path=pdf_path)



from musurgia.agpdf.textlabel import TextLabel


class Labeled(object):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._text_labels = []

    def add_text_label(self, val, **kwargs):
        if val is not None and not isinstance(val, TextLabel):
            val = TextLabel(val, **kwargs)
        val.parent = self
        self._text_labels.append(val)
        return val

    @property
    def text_labels(self):
        return self._text_labels

    def remove_text_labels(self):
        self._text_labels = []



from musurgia.agpdf.drawobject import DrawObject
from musurgia.agpdf.labeled import Labeled
from musurgia.agpdf.line import Line
from musurgia.agpdf.named import Named


class SegmentedLine(DrawObject, Labeled, Named):
    def __init__(self, lengths, factor=1, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._lines = []
        self._factor = None
        self.factor = factor
        self._lengths = None
        self.lengths = lengths

    @DrawObject.relative_y.setter
    def relative_y(self, val):
        self._relative_y = val
        try:
            for line in self.lines:
                line.relative_y = val
        except AttributeError:
            pass

    @DrawObject.relative_x.setter
    def relative_x(self, val):
        self._relative_x = val
        try:
            self.lines[-1].relative_x = val
        except AttributeError:
            pass

    @property
    def lengths(self):
        return self._lengths

    @lengths.setter
    def lengths(self, val):
        if not isinstance(val, list):
            raise TypeError('lengths.value must be of type list not{}'.format(type(val)))
        self._lengths = val
        self._generate_lines()

    @property
    def lines(self):
        return self._lines

    def _generate_lines(self):
        self._lines = []
        for length in self.lengths:
            line = Line(length=length, relative_y=self.relative_y, factor=self.factor)
            if not self._lines:
                line.relative_x = self.relative_x
            else:
                line.relative_x = 0

            self._lines.append(line)
        self._lines[-1].end_mark_line.show = True

    def get_relative_x2(self):
        raise Exception('SegementedLine does not have a x2 value')

    def get_relative_y2(self):
        return self.lines[0].relative_y

    def draw(self, pdf):
        for text_label in self._text_labels:
            text_label.draw(pdf)

        for line in self.lines:
            if line == self.lines[0]:
                line.name = self.name

            line.draw_with_break(pdf)
            new_x = pdf.x

            if line._line_break and self.name:
                line.name = self.name
                pdf.x = new_x - line.actual_length
                line.name.draw(pdf)
                pdf.x = new_x



import os

from musurgia.agpdf.textlabel import Text
from fpdf import FPDF


class PageText(Text):
    def __init__(self, text, v_position=None, h_position=None, *args, **kwargs):
        super().__init__(text=text, *args, **kwargs)
        self.v_position = v_position
        self.h_position = h_position

    def get_text_physical_length(self):
        return (len(self.text)) * self.font_size / 6

    def draw(self, pdf):
        old_x, old_y = pdf.x, pdf.y
        if self.v_position == 'center':
            pdf.x = (pdf.w / 2) - self.get_text_physical_length() / 2
        elif self.v_position == 'left':
            pdf.x = pdf.l_margin
        elif self.v_position == 'right':
            pdf.x = pdf.w - pdf.r_margin - self.get_text_physical_length()
        else:
            pass

        if self.h_position == 'top':
            pdf.y = pdf.t_margin
        elif self.h_position == 'bottom':
            pdf.y = pdf.h - pdf.b_margin
        else:
            pass
        super().draw(pdf)
        pdf.x, pdf.y = old_x, old_y


class PageNumber(PageText):
    def __init__(self, text='none', v_position='center', h_position='bottom', *args, **kwargs):
        super().__init__(text=text, v_position=v_position, h_position=h_position, *args, **kwargs)

    def __call__(self, val):
        self.text = val
        self.page = val


class Pdf(FPDF):

    def __init__(self, default_line_distance=10, r_margin=10, t_margin=10, l_margin=10, b_margin=10, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.page_number = PageNumber('')
        self.r_margin = r_margin
        self.t_margin = t_margin
        self.l_margin = l_margin
        self.b_margin = b_margin
        self.add_page()

        self.default_line_distance = default_line_distance
        self.set_font("Arial", "", 10)

    def add_page(self):
        super().add_page(orientation=self.cur_orientation)

    def draw_page_number(self):
        for page in self.pages:
            self.page = page
            self.page_number(page)
            self.page_number.draw(self)

    def add_space(self, val):
        self.y += val

    def write(self, path):
        self.draw_page_number()
        self.output(path, 'F')

    def show(self, path):
        self.write(path)
        os.system('open ' + path)



from quicktions import Fraction


class InterpolationSection(object):
    def __init__(self, start, end, duration, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._start = None
        self._end = None
        self._duration = None
        self.start = start
        self.end = end
        self.duration = duration

    def get_value(self, x):
        if x < 0 or x > self.duration:
            raise ValueError()
        return Fraction(Fraction(x * (self.end - self.start)), self.duration) + self.start


class Interpolation(object):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._sections = []

    def add_section(self, start, end, duration):
        section = InterpolationSection(start, end, duration)

        if not isinstance(section, InterpolationSection):
            raise TypeError()
        self._sections.append(section)

    def get_value(self, x):
        temp_x = x
        for section in self._sections:
            try:
                return section.get_value(temp_x)
            except ValueError:
                temp_x -= section.duration

        raise ValueError(x)



import os
from unittest import TestCase

from musurgia.fractaltree.fractalmusic import FractalMusic
from musurgia.testcomparefiles import TestCompareFiles

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def setUp(self) -> None:
        self.fm = FractalMusic(proportions=[1, 2, 3, 4], tree_permutation_order=[3, 1, 4, 2], duration=10.4)

    def test_1(self):
        self.fm.tempo = self.fm.find_best_tempo()
        self.assertEqual(self.fm.duration, 10.4)

    def test_2(self):
        self.fm.tempo = self.fm.find_best_tempo()
        self.assertEqual(self.fm.quarter_duration, 13)

    def test_3(self):
        self.fm.add_layer()
        for child in self.fm.get_children():
            best_tempo = child.find_best_tempo()
            child.tempo = best_tempo
        # print([leaf.quarter_duration for leaf in self.fm.get_children()])
        self.fm.round_leaves()
        result = [4.0, 1.0, 7.0, 2.0]
        self.assertEqual([child.quarter_duration for child in self.fm.get_children()], result)

    def test_4(self):
        self.fm.add_layer()
        for child in self.fm.get_children():
            best_tempo = child.find_best_tempo()
            child.tempo = best_tempo
        self.fm.round_leaves()
        xml_path = path + '_test_4.xml'
        score = self.fm.get_root_score()
        score.write(xml_path)
        TestCompareFiles().assertTemplate(xml_path)



import os
from time import sleep
from unittest import TestCase

from musurgia.agpdf.pdf import Pdf
from musurgia.agpdf.segmentedline import SegmentedLine
from musurgia.agpdf.textlabel import TextLabel
from musurgia.testcomparefiles import TestCompareFiles

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def test_1(self):
        pdf_path = path + '_test_1.pdf'
        sl = SegmentedLine(lengths=[1, 3, 2, 5], factor=5)
        pdf = Pdf('portrait')
        sl.draw(pdf)
        pdf.write(pdf_path)
        TestCompareFiles().assertTemplate(file_path=pdf_path)

    def test_2(self):
        pdf_path = path + '_test_2.pdf'
        sl = SegmentedLine(lengths=30 * [20])
        pdf = Pdf('portrait')
        sl.draw(pdf)
        pdf.write(pdf_path)
        TestCompareFiles().assertTemplate(file_path=pdf_path)

    def test_3(self):
        pdf_path = path + '_test_3.pdf'
        sl = SegmentedLine(lengths=200 * [20])
        pdf = Pdf('portrait')
        sl.draw(pdf)
        pdf.write(pdf_path)
        TestCompareFiles().assertTemplate(file_path=pdf_path)

    def test_4(self):
        pdf_path = path + '_test_4.pdf'
        sl = SegmentedLine(lengths=30 * [10])
        sl.add_text_label(TextLabel('bla is bla'))
        pdf = Pdf('portrait')
        sl.draw(pdf)
        pdf.write(pdf_path)
        TestCompareFiles().assertTemplate(file_path=pdf_path)

    def test_5(self):
        pdf_path = path + '_test_5.pdf'
        sl = SegmentedLine(lengths=30 * [10], name='vla')
        pdf = Pdf('portrait')
        sl.draw(pdf)
        # pdf.add_draw_object(sl)
        pdf.write(pdf_path)
        TestCompareFiles().assertTemplate(file_path=pdf_path)
